#include <stdio.h>
#include <stdlib.h>
#include <alsa/asoundlib.h>
#include <stdint.h>
#include <time.h>
#include "alsa.h"
#include "audio_buffer.h"
#include "thread.h"
#include "download.h"
#include "utils.h"
#include "eq.h"
#include "conf.h"
#include "logger.h"

extern volatile int do_eq;

#define A do { if (err < 0) { logprintf("alsa-out-xrun: %s:%d: alsa error '%s'\n", __FILE__, __LINE__, snd_strerror(err)); abort(); } } while (0)

static void open_audio(alsa_data *a)
{
  snd_pcm_t           *h;
  snd_pcm_hw_params_t *hw_params;
  snd_pcm_sw_params_t *sw_params;
  snd_pcm_uframes_t   period_frames;
  snd_pcm_uframes_t   buffer_frames;
  unsigned int        rate;
  snd_pcm_format_t    format;
  int err;
  format = SND_PCM_FORMAT_S16_LE;
  err = snd_pcm_open(&h, a->device, SND_PCM_STREAM_PLAYBACK, 0); A;
  snd_pcm_hw_params_alloca(&hw_params);
  snd_pcm_sw_params_alloca(&sw_params);
  err = snd_pcm_hw_params_any(h, hw_params); A;
  err = snd_pcm_hw_params_set_access(h, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED); A;
  err = snd_pcm_hw_params_set_format(h, hw_params, format); A;
  err = snd_pcm_hw_params_set_channels(h, hw_params, a->channels); A;
  rate = a->samplerate;
  err = snd_pcm_hw_params_set_rate_near(h, hw_params, &rate, 0); A;
  if (rate != a->samplerate) {logprintf("bad rate\n"); abort();}
  /* dumb values */
  period_frames = 1024; //a->nsamples;
  buffer_frames = 10 * 1024;   //period_frames * a->nperiods;
  err = snd_pcm_hw_params_set_period_size_near(h,hw_params,&period_frames,0);A;
  err = snd_pcm_hw_params_set_buffer_size_near(h, hw_params, &buffer_frames);A;
  err = snd_pcm_hw_params(h, hw_params); A;
//  err = snd_pcm_start(h); A;
  a->h = h;
}

static void xrun(alsa_data *a)
{
  char data[1024*2*2];
  logprintf("xrun\n");
  memset(data, 0, 1024*2*2);
  snd_pcm_prepare(a->h);
  snd_pcm_writei(a->h, data, 1024);
}

static void suspend(alsa_data *a)
{
  int res;
  int loop = 0;
  logprintf("suspend\n");
  while ((res = snd_pcm_resume(a->h)) == -EAGAIN && loop < 100) {
    struct timeval t;
    t.tv_sec = 0;
    t.tv_usec = 10 * 1000;
    select(0, 0, 0, 0, &t);
    loop++;
  }
  if (res < 0) xrun(a); //snd_pcm_prepare(a->h);
}

static void put_audio_alsa(alsa_data *a, char *buf, int nsamples)
{
#define X do { logprintf("alsa-out-xrun: %s:%d: alsa error\n", __FILE__, __LINE__); xrun(a); } while (0)
#define Y do { logprintf("alsa-out-xrun: %s:%d: alsa error\n", __FILE__, __LINE__); suspend(a); } while (0)
  size_t count = nsamples;
  ssize_t r;
  unsigned char *bb;

  /* your compiler might go mad here */
  {
    short *a = (short*)buf;
    float *b = (float*)buf;
    int sp = nsamples * 2;
    int i;
    for (i = 0; i < sp; i++) {
      float v = b[i] * 32768;
      if (v > 32767) v = 32767;
      if (v < -32768) v = -32768;
      a[i] = v;
    }
  }

  bb = (unsigned char *)buf;

  while (count) {
    r = snd_pcm_writei(a->h, bb, count);
    if (r == -EAGAIN || (r >= 0 && (size_t)r < count)) {
      /* nothing */
    }
    else if (r == -EPIPE) X;
    else if (r == -ESTRPIPE) Y;
    else if (r < 0) X;
    if (r > 0) {
      count -= r;
      bb += r * a->channels * 2;
    }
  }
#undef X
}

volatile uint64_t audio_start_time;

eq e;

static audio_item block_20ms;
static audio_item block_current;
static int block_next_sample;
static int block_samples;

/* speed that thing up (memmove & co.) */
static void get_audio_20ms(int samples_per_20ms, int samplesize)
{
  int i;

  i = 0;

next_block:
  if (block_next_sample != block_samples) {
    for (; i < samples_per_20ms && block_next_sample != block_samples;
         i++, block_next_sample++) {
      int j;
      for (j = 0; j < samplesize; j++)
        block_20ms.data[i * samplesize + j] =
            block_current.data[block_next_sample * samplesize + j];
    }
  }

  if (i == samples_per_20ms) return;

  free(block_current.data);
  block_current = get_audio();
  block_samples = block_current.count / samplesize;
  block_next_sample = 0;
  if (its_over) return;
  goto next_block;
}

static void *alsa_thread(void *_a)
{
  alsa_data *a = _a;
  uint64_t frame_streamtime;
  uint64_t frame_consumed = 0;
  uint64_t world_start_time = get_start_time();
  uint64_t video_start_time = get_video_start_time();
  uint64_t next_play_time;
  uint64_t cur_time;
  uint64_t next_image_streamtime;
  struct timespec cur_time_ts;
  snd_pcm_sframes_t fr;
  char *silence;
  int silence_samples;
  int samples_per_20ms;
  int ret;
  int i;
  float *audio_data;
  int audio_samples;
  extern volatile float volume;
  extern volatile uint64_t audio_delay;
  int framesize_ms;

  if (a->channels < 1 || a->channels > 2) abort();
  if (a->samplerate < 8000 || a->samplerate > 48000) abort();

  init_eq(&e, a->samplerate);

  e.filter = 1;

  samples_per_20ms = 20 * a->samplerate / 1000;
  /* get 'multiple of 2' number of samples */
  samples_per_20ms++;
  samples_per_20ms &= ~1;

  silence_samples = samples_per_20ms * 2;
  silence = malloc(silence_samples * a->channels * sizeof(float));
  if (silence == NULL) abort();

  block_20ms.data = malloc(samples_per_20ms * a->channels * sizeof(float));
  if (block_20ms.data == NULL) abort();
  block_20ms.count = samples_per_20ms * a->channels * sizeof(float);

  audio_start_time = get_audio_start_time();

  a->device = "default";

  open_audio(a);

  while (!its_over) {
    get_audio_20ms(samples_per_20ms, a->channels * sizeof(float));
    if (its_over) goto quit;
    frame_streamtime = audio_start_time + audio_delay +
                       (uint64_t)frame_consumed * 1000 / a->samplerate;
    frame_consumed += samples_per_20ms;
    ret = snd_pcm_delay(a->h, &fr);
    if (ret) fr = 0;
    /* useless test? can it be <0? */
    if (fr < 0) fr = 0;
    if (clock_gettime(CLOCK_MONOTONIC_RAW, &cur_time_ts)) abort();
    cur_time = cur_time_ts.tv_sec * 1000 + cur_time_ts.tv_nsec / 1000000;
    next_play_time = cur_time + (uint64_t)fr * 1000 / a->samplerate;
    next_image_streamtime = video_start_time +
                            (next_play_time - world_start_time);
    framesize_ms = samples_per_20ms * 1000 / a->samplerate;
    if (frame_streamtime < next_image_streamtime-framesize_ms*2) {
      /* too late, play at speed*2 */
      float *src = (float *)block_20ms.data;
      float *dst = (float *)silence;
      for (i = 0; i < samples_per_20ms/2; i++) {
        int j;
        for (j = 0; j < a->channels; j++)
          dst[j] = (src[j] + src[j+a->channels])/2;
        dst += a->channels;
        src += a->channels * 2;
      }
      audio_data = (float *)silence;
      audio_samples = samples_per_20ms / 2;
    } else
    if (frame_streamtime > next_image_streamtime+framesize_ms*2) {
      /* too early, play at speed/2 */
      float *src = (float *)block_20ms.data;
      float *dst = (float *)silence;
      for (i = 0; i < samples_per_20ms; i++) {
        int j;
        for (j = 0; j < a->channels; j++) dst[j] = src[j];
        dst += a->channels;
        for (j = 0; j < a->channels; j++) dst[j] = src[j];
        dst += a->channels;
        src += a->channels;
      }
      audio_data = (float *)silence;
      audio_samples = samples_per_20ms * 2;
    } else {
      /* normal speed */
      audio_data = (float *)block_20ms.data;
      audio_samples = samples_per_20ms;
    }
    for (i = 0; i < audio_samples * a->channels; i++) audio_data[i] *= volume;
    if (do_eq) run_eq(&e, audio_data, audio_samples, a->channels);
    put_audio_alsa(a, (char*)audio_data, audio_samples);
    logprintf("ALSA delay %ld\n", fr);
  }

quit:
  if (snd_pcm_drop(a->h)) abort();
  if (snd_pcm_close(a->h)) abort();
  free(block_current.data);
  block_current.data = NULL;
  block_current.count = 0;
  block_samples = 0;
  block_next_sample = 0;
  free(block_20ms.data);
  block_20ms.data = NULL;
  block_20ms.count = 0;
  free(silence);
  close_eq(&e);

  return NULL;
}

static pthread_t ath;

void alsa_init(alsa_data *d)
{
  ath = new_thread(alsa_thread, d);
}

void alsa_close(void)
{
  void *tret;
  if (pthread_join(ath, &tret)) abort();
}
