#include "audio_buffer.h"
#include "utils.h"
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "logger.h"

static audio_buffer buffer;

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

void put_audio(char *audio, int count)
{
  audio_item *n;

  if (pthread_mutex_lock(&mutex)) abort();

  n = malloc(sizeof(*n)); if (n == NULL) abort();
  n->data = malloc(count); if (n->data == NULL) abort();
  memcpy(n->data, audio, count);
  n->count = count;
  n->next = NULL;
  if (buffer.end != NULL) buffer.end->next = n;
  if (buffer.start == NULL) buffer.start = n;
  buffer.end = n;

  buffer.count++;

  if (pthread_cond_broadcast(&cond)) abort();
  if (pthread_mutex_unlock(&mutex)) abort();
}

audio_item get_audio(void)
{
  audio_item ret;
  audio_item *t;

  if (pthread_mutex_lock(&mutex)) abort();
  if (its_over) goto over;
  while (buffer.start == NULL) {
    if (pthread_cond_wait(&cond, &mutex)) abort();
    if (its_over) {
over:
      ret.data = NULL; ret.count = 0;
      if (pthread_mutex_unlock(&mutex)) abort();
      return ret;
    }
  }

  ret = *buffer.start;

  t = buffer.start;
  buffer.start = t->next;
  if (t == buffer.end) buffer.end = NULL;
  free(t);

  buffer.count--;
  logprintf("audio buffer count %d\n", buffer.count);

  if (pthread_cond_broadcast(&cond)) abort();
  if (pthread_mutex_unlock(&mutex)) abort();

  return ret;
}

void its_over_buffer_audio(void)
{
  if (pthread_mutex_lock(&mutex)) abort();
  if (pthread_cond_broadcast(&cond)) abort();
  if (pthread_mutex_unlock(&mutex)) abort();
}

void reset_audio_buffer(void)
{
  while (buffer.start) {
    audio_item *it = buffer.start;
    free(it->data);
    buffer.start = buffer.start->next;
  }

  buffer.start = NULL;
  buffer.end = NULL;
  buffer.count = 0;
}
