#include "jsmn.c"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

int get_element(char *s, jsmntok_t *t, int dad, char *elem)
{
  int i;
  int cur = dad+1;
  if (t[dad].type != JSMN_OBJECT) return -1;

  for (i = 0; i < t[dad].size; i++) {
    if (t[cur].type != JSMN_STRING) return -1;
    if (!strncmp(elem, s+t[cur].start, t[cur].end-t[cur].start))
      return cur+1;
    cur++;
    if (i != t[dad].size - 1) while (t[cur].parent != dad) cur++;
  }

  return -1;
}

int get_array(jsmntok_t *t, int dad, int index)
{
  int i;
  int cur = dad+1;
  for (i = 0; i < index; i++) {
    cur++;
    while (t[cur].parent != dad) cur++;
  }
  return cur;
}

void dump_programs(char *s, jsmntok_t *t, int item)
{
  int i;
  int p = get_element(s, t, item, "programs");
  if (p == -1) return;
  if (t[p].type != JSMN_ARRAY) return;
  for (i = 0; i < t[p].size; i++) {
    uint64_t dd;
    time_t dt;
    struct tm *dtm;
    int hstart, mstart, hend, mend;
    int prog = get_array(t, p, i);
    int title, date, duration, genre, /*synopsis,*/ summary, additional_data;
    if (t[prog].type != JSMN_OBJECT) return;
    title = get_element(s, t, prog, "title");
    date = get_element(s, t, prog, "diffusionDate");
    duration = get_element(s, t, prog, "duration");
    genre = get_element(s, t, prog, "genre");
    //synopsis = get_element(s, t, prog, "synopsis");
    additional_data = get_element(s, t, prog, "additionnalData");
    summary = get_element(s, t, additional_data, "longSummary");
    dd = atoll(s+t[date].start) / 1000;
    dt = dd;
    dtm = localtime(&dt);
    hstart = dtm->tm_hour;
    mstart = dtm->tm_min;
    dd = atoll(s+t[duration].start);
    dt += dd * 60;
    dtm = localtime(&dt);
    hend = dtm->tm_hour;
    mend = dtm->tm_min;
#if 0
    printf("[%2.2dh%2.2d - %2.2dh%2.2d] %.*s        [%.*s]\n",
           hstart, mstart, hend, mend,
           t[title].end-t[title].start, s+t[title].start,
           t[genre].end-t[genre].start, s+t[genre].start);
#endif
    printf("[%2.2dh%2.2d - %2.2dh%2.2d] %.*s\n",
           hstart, mstart, hend, mend,
           t[title].end-t[title].start, s+t[title].start);
    //printf("résumé: %.*s\n", t[synopsis].end-t[synopsis].start, s+t[synopsis].start);
    //printf("résumé long: %.*s\n", t[summary].end-t[summary].start, s+t[summary].start);
  }
  printf("ok\n");
}

int main(void)
{
  jsmn_parser p;
  jsmntok_t *tokens;
  char s[649451];
  FILE *f;f=fopen("./servicePlan.json","r");fread(s,649451,1,f);fclose(f);
  jsmnerr_t r;
  int response;
  int i;

  jsmn_init(&p);
  r = jsmn_parse(&p, s, 649451, NULL, 0);
  if (r < 0) { printf("parse error\n"); return 1; }
  tokens = malloc(r * sizeof(jsmntok_t)); if (tokens == NULL) abort();
  jsmn_init(&p);
  r = jsmn_parse(&p, s, 649451, tokens, r);
  if (r < 0) { printf("parse error\n"); return 1; }

  if (tokens[0].type != JSMN_OBJECT)
    { printf("root not object\n"); return 1; }

  response = get_element(s, tokens, 0, "response");
  if (response == -1 || tokens[response].type != JSMN_ARRAY)
    { printf("no response item or bad type\n"); return 1; }
  for (i = 0; i < tokens[response].size; i++) {
    int item = get_array(tokens, response, i);
    int name = get_element(s, tokens, item, "name");
    int id   = get_element(s, tokens, item, "channelId");
    int drm  = get_element(s, tokens, item, "drmized");
    printf("kid %d: %d name %.*s id %.*s drm %.*s\n", i, item,
           tokens[name].end-tokens[name].start, s+tokens[name].start,
           tokens[id].end-tokens[id].start, s+tokens[id].start,
           tokens[drm].end-tokens[drm].start, s+tokens[drm].start);
    dump_programs(s, tokens, item);
  }

  char *t =  "{\"a\":{\"suba0\":0, \"suba1\":1 }, \"b\": 2 }";
  jsmn_init(&p);
  r = jsmn_parse(&p, t, strlen(t), tokens, r);

  return 0;
}
