#include <string.h>
#include <stdio.h>
#include <setjmp.h>
#include <stdint.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <inttypes.h>
#include "download.h"
#include "utils.h"
#include "logger.h"

//abcdXXXefgh
//XXX
//YY
void replace(char *s, char *t, char *nt)
{
  int nt_len = strlen(nt);
  int t_len = strlen(t);
  int s_len = strlen(s);
  char *u = strstr(s, t); if (u == NULL) return;
  memmove(u + nt_len, u + t_len, s_len - (u-s) - t_len);
  memmove(u, nt, nt_len);
  s[s_len - t_len + nt_len] = 0; 
}

typedef struct {
  char *data;
  int size;
  int pos;
} boxbuf;

static uint32_t getu32(boxbuf *b, jmp_buf j)
{
  uint32_t r;
  if (b->size - b->pos < 4) longjmp(j, 1);
  r = ((unsigned char)b->data[b->pos] << 24) |
      ((unsigned char)b->data[b->pos+1] << 16) |
      ((unsigned char)b->data[b->pos+2] << 8) |
      ((unsigned char)b->data[b->pos+3]);
  b->pos += 4;
  return r;
}

static uint64_t getu64(boxbuf *b, jmp_buf j)
{
  uint64_t r;
  if (b->size - b->pos < 8) longjmp(j, 1);
  r = ((uint64_t)((unsigned char)b->data[b->pos]) << (32+24)) |
      ((uint64_t)((unsigned char)b->data[b->pos+1]) << (32+16)) |
      ((uint64_t)((unsigned char)b->data[b->pos+2]) << (32+8)) |
      ((uint64_t)((unsigned char)b->data[b->pos+3]) << 32) |
      ((uint64_t)((unsigned char)b->data[b->pos+4]) << 24) |
      ((uint64_t)((unsigned char)b->data[b->pos+5]) << 16) |
      ((uint64_t)((unsigned char)b->data[b->pos+6]) << 8) |
      ((uint64_t)((unsigned char)b->data[b->pos+7]));
  b->pos += 8;
  return r;
}

void get_mdat(dbuf *in, dbuf *out)
{
  boxbuf b;
  jmp_buf j;

  b.data = in->data;
  b.size = in->size;
  b.pos = 0;

  out->error = 0;

  if (setjmp(j)) {
    out->error = 1;
    logprintf("error with audio data\n");
    return;
  }

  while (b.pos < b.size) {
    int dapos = b.pos;
    uint64_t boxsize = getu32(&b, j);
    uint32_t boxtype = getu32(&b, j);
    if (boxsize == 1) boxsize = getu64(&b, j);
    if (boxsize == 0) boxsize = b.size - dapos;
    logprintf("box %c%c%c%c size %"PRIu64"\n",
        (boxtype >> 24) & 255,
        (boxtype >> 16) & 255,
        (boxtype >> 8) & 255,
        (boxtype) & 255, boxsize);
    if (boxtype == 0x6d646174) {
      out->data = b.data + b.pos;
      out->size = boxsize-(b.pos-dapos);
      return;
    }
    b.pos = dapos + boxsize;
  }
}

int hex2raw(char *out, char *in)
{
  int ret = 0;
  while (*in) {
    int a = (unsigned char)in[0];
    int b = (unsigned char)in[1];
    in += 2;
    if (a >= '0' && a <= '9') a = a - '0';
    else if (a >= 'A' && a <= 'F') a = 10 + a - 'A';
    else if (a >= 'a' && a <= 'f') a = 10 + a - 'a';
    else abort();
    if (b >= '0' && b <= '9') b = b - '0';
    else if (b >= 'A' && b <= 'F') b = 10 + b - 'A';
    else if (b >= 'a' && b <= 'f') b = 10 + b - 'a';
    else abort();
    out[ret] = (a<<4) | b;
    ret++;
  }
  return ret;
}

static uint64_t _start_time;
static uint64_t _video_start_time;
static uint64_t _audio_start_time;

void set_start_time(void)
{
  struct timespec start_time_ts;
  if (clock_gettime(CLOCK_MONOTONIC_RAW, &start_time_ts)) abort();
  _start_time = (uint64_t)start_time_ts.tv_sec * 1000 +
                (uint64_t)start_time_ts.tv_nsec / 1000000;
}

uint64_t get_start_time(void)
{
  return _start_time;
}

void set_video_start_time(uint64_t t)
{
  _video_start_time = t;
}

uint64_t get_video_start_time(void)
{
  return _video_start_time;
}

void set_audio_start_time(uint64_t t)
{
  _audio_start_time = t;
}

uint64_t get_audio_start_time(void)
{
  return _audio_start_time;
}

int volatile its_over;

int get_utf8(char *s, int *v)
{
#define V do { *v = *n; return 1; } while (0)
  int len;
  int val;
  unsigned char *n = (unsigned char *)s;

  if (*n == 0xc0 || *n == 0xc1 || *n >= 0xf5) V;
  if (!(*n & 0x80)) V;
  else
  if ((*n & 0xe0) == 0xc0) { /* 110xxxxx */
    if ((*(n+1) & 0xc0) != 0x80) V;
    val = ((*n & 0x1f) << 6) | (*(n+1) & 0x3f);
    if (val < 0x80 || val > 0x7ff) V;
    *v = val;
    len = 2;
  } else
  if ((*n & 0xf0) == 0xe0) { /* 1110xxxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80) V;
    val = ((*n & 0x0f) << (6*2)) | ((*(n+1) & 0x3f) << 6) | (*(n+2) & 0x3f);
    if (val < 0x800 || val > 0xffff) V;
    *v = val;
    len = 3;
  } else
  if ((*n & 0xf8) == 0xf0) { /* 11110xxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80 ||
        (*(n+3) & 0xc0) != 0x80) V;
    val = ((*n & 0x07) << (6*3)) | ((*(n+1) & 0x3f) << (6*2)) |
          ((*(n+2) & 0x3f) << 6) | (*(n+3) & 0x3f);
    if (val < 0x10000 || val > 0x10ffff) V;
    *v = val;
    len = 4;
  } else V;
  /* we don't deal with values > 0xffff */
  if (val > 0xffff) V;
  return len;
#undef V
}

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

void lock_channels_list(void)
{
  if (pthread_mutex_lock(&lock)) abort();
}

void unlock_channels_list(void)
{
  if (pthread_mutex_unlock(&lock)) abort();
}
