#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <inttypes.h>
#include <sys/select.h>
#include "video_buffer.h"
#include "download.h"
#include "utils.h"
#include "text_over.h"
#include "guide.h"
#include "eqgui.h"
#include "logger.h"

#define GUI_MODE_DEFAULT    0
#define GUI_MODE_GUIDE      1
#define GUI_MODE_FULL_GUIDE 2
#define GUI_MODE_EQ         3
#define GUI_MODE_HELP       4

static Display *d;
static Window w;
static GLXContext glc;
static int fullscreen;
static XVisualInfo *vi;
static GLuint texture[2];

volatile float volume = 1;
static float volumeDB = 0;
static int mute = 0;
static float aspect_ratio_adjust = 1;

volatile int do_eq = 1;

volatile uint64_t audio_delay;

extern void next_channel(int direction);
extern int set_channel(int channel);

volatile int x_can_go;

static void select_input(void)
{
  XSelectInput(d, w, KeyPressMask | EnterWindowMask | PointerMotionMask |
               StructureNotifyMask | ExposureMask | ButtonPressMask |
               ButtonReleaseMask);
}

static void init(int width, int h)
{
  GLint att[] = { GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None };

//  static int inited = 0;
//  if (inited) return;
//  inited = 1;
  d = XOpenDisplay(0); if (d==NULL) abort();

  vi = glXChooseVisual(d, 0, att);
  if (vi == NULL) { fprintf(stderr, "gl error\n"); abort(); }

  w = XCreateWindow(d, DefaultRootWindow(d),
          0, 0, width, h, 0, vi->depth, InputOutput, vi->visual, 0, NULL);

  select_input();
  XMapWindow(d, w);

  glc = glXCreateContext(d, vi, NULL, GL_TRUE);
  glXMakeCurrent(d, w, glc);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glGenTextures(2, texture);
  glEnable(GL_TEXTURE_2D);
  //glActiveTexture(GL_TEXTURE0+0);
  glBindTexture(GL_TEXTURE_2D, texture[0]);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glBindTexture(GL_TEXTURE_2D, texture[1]);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

  //XFlush(d);
}

/* this code may not be the best, even may be wrong */
/* if you know opengl maybe you can tell me... */
void display_image(id_image image, int width, int height,
                   int window_width, int window_height,
                   eqgui *eg)
{
  static int last_width = -1, last_height = -1;
  static int last_id = -1;
  static int last_text_width = -1, last_text_height = -1;
  static int last_over_id = -1;
  static int last_over = -1;  /* 0: last was text_over, 1: eq, -1: nothing */
  double scale;
  double sw, sh;
  id_image text_over;
  int current_over;

  if (eg->visible) {
    text_over = eqgui_get_screen(eg);
    current_over = 1;
  } else {
    text_over = get_text_over();
    current_over = 0;
  }

  glViewport(0, 0, window_width, window_height);
  glClear(GL_COLOR_BUFFER_BIT);

  if (image.data == NULL && text_over.data == NULL) {
    glXSwapBuffers(d, w);
    return;
  }

  if (image.data == NULL) goto no_image;

  scale = window_width / ((double)width*aspect_ratio_adjust);
  if (scale * height > window_height)
    scale = window_height / (double)height;
  sw = scale*width*aspect_ratio_adjust;
  sh = scale*height;
  glViewport((window_width-sw)/2, (window_height-sh)/2, sw, sh);
  //glActiveTexture(GL_TEXTURE0+0);
  glBindTexture(GL_TEXTURE_2D, texture[0]);
  if (width != last_width || height != last_height) {
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, image.data);
    last_width = width;
    last_height = height;
  } else if (last_id != image.id)
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height,
                    GL_RGBA, GL_UNSIGNED_BYTE, image.data);
  glLoadIdentity();
  glBegin(GL_QUADS);
  glTexCoord2f(1, 1); glVertex3f( 1, -1, 0);
  glTexCoord2f(0, 1); glVertex3f(-1, -1, 0);
  glTexCoord2f(0, 0); glVertex3f(-1,  1, 0);
  glTexCoord2f(1, 0); glVertex3f( 1,  1, 0);

  glEnd();                          // Done Drawing The Background
  last_id = image.id;

  //free(image);

no_image:
  if (text_over.data == NULL) goto no_text_over;

  width = 640;
  height = 352;

  scale = window_width / (double)width;
  if (scale * height > window_height)
    scale = window_height / (double)height;
  sw = scale*width;
  sh = scale*height;
  glViewport((window_width-sw)/2, (window_height-sh)/2, sw, sh);
  //glViewport(0, 0, 100, 100);
  //glActiveTexture(GL_TEXTURE0+1);
  glBindTexture(GL_TEXTURE_2D, texture[1]);
  if (width != last_text_width || height != last_text_height) {
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, text_over.data);
    last_text_width = width;
    last_text_height = height;
  } else if (current_over != last_over || text_over.id != last_over_id)
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height,
                    GL_RGBA, GL_UNSIGNED_BYTE, text_over.data);
  glLoadIdentity();
  glBegin(GL_QUADS);
  glTexCoord2f(1, 1); glVertex3f( 1, -1, 0);
  glTexCoord2f(0, 1); glVertex3f(-1, -1, 0);
  glTexCoord2f(0, 0); glVertex3f(-1,  1, 0);
  glTexCoord2f(1, 0); glVertex3f( 1,  1, 0);
  glEnd();

  last_over_id = text_over.id;
  last_over = current_over;

no_text_over:
  glXSwapBuffers(d, w);
}

static void toggle_fullscreen(int last_x, int last_y,
    int last_width, int last_height)
{
  /* important: detach old GL context before destroying window */
  glXMakeCurrent(d, None, NULL);

  /* that works for me (fvwm), it might well fail for you */
  if (fullscreen) {
    XSizeHints hints;
    XDestroyWindow(d, w);
    w = XCreateWindow(d, DefaultRootWindow(d), 0, 0,
                      last_width, last_height, 0,
                      vi->depth, InputOutput, vi->visual, 0, NULL);
    hints.flags = USPosition;
    hints.x = last_x;
    hints.y = last_y;
    XSetWMNormalHints(d, w, &hints);
    select_input();
    XMapWindow(d, w);
    XMoveWindow(d, w, last_x, last_y);
    fullscreen = 0;
  } else {
    Atom wm_state;
    Atom wm_fullscreen;
    XSizeHints hints;
    XWindowAttributes wa;
    XGetWindowAttributes(d, DefaultRootWindow(d), &wa);
    XDestroyWindow(d, w);
    w = XCreateWindow(d, DefaultRootWindow(d), 0, 0, wa.width, wa.height, 0,
                      vi->depth, InputOutput, vi->visual, 0, NULL);
    wm_state   = XInternAtom (d, "_NET_WM_STATE", True );
    wm_fullscreen = XInternAtom (d, "_NET_WM_STATE_FULLSCREEN", True );
    XChangeProperty(d, w, wm_state, XA_ATOM, 32,
                    PropModeReplace, (unsigned char *)&wm_fullscreen, 1);
//    XMoveWindow(d, w, 0, 0);
    hints.flags = USPosition;
    hints.x = hints.y = 0;
    XSetWMNormalHints(d, w, &hints);
    select_input();
    XMapWindow(d, w);
    fullscreen = 1;
  }
  glXMakeCurrent(d, w, glc);
  //XFlush(d);
}

static void set_window_size_from_video(int width, int height)
{
  glXMakeCurrent(d, None, NULL);
  XDestroyWindow(d, w);

  w = XCreateWindow(d, DefaultRootWindow(d), 0, 0, width, height, 0,
                    vi->depth, InputOutput, vi->visual, 0, NULL);
  XMapWindow(d, w);
  fullscreen = 0;
  glXMakeCurrent(d, w, glc);
  select_input();
  //XFlush(d);
}

void hide_pointer(void)
{
  Cursor invisibleCursor;
  Pixmap bitmapNoData;
  XColor black;
  static char noData[] = { 0,0,0,0,0,0,0,0 };
  black.red = black.green = black.blue = 0;

  bitmapNoData = XCreateBitmapFromData(d, w, noData, 8, 8);
  invisibleCursor = XCreatePixmapCursor(d, bitmapNoData, bitmapNoData, 
                                       &black, &black, 0, 0);
  XDefineCursor(d, w, invisibleCursor);
  XFreeCursor(d, invisibleCursor);
  XFreePixmap(d, bitmapNoData);
}

void show_pointer(void)
{
  XUndefineCursor(d, w);
}

void *xthread(void *_)
{
  int last_x = 0;
  int last_y = 0;
  int last_width;
  int last_height;
  int display_width = 640;
  int display_height = 360;
  image_item current_image = { NULL, 0, 0 };
  image_item next_image = { NULL, 0, 0 };
  struct timespec cur_time;
  int image_number;
  unsigned long cur_time_ms, next_time_ms, nt_ms;
  unsigned long delay_ms;
  int pointer_visible = 1;
  unsigned long last_pointer_time;
  //extern volatile uint64_t audio_start_time;
  //uint64_t audio_start_time_orig;
  float aspect_ratios[] = { 0, 4./3, 16./9 };
  char *aspect_ratios_name[] = { "default", "4/3", "16/9" };
  int current_aspect_ratio = 0;
  fd_set rset;
  struct timeval timeout;
  int gui_mode = GUI_MODE_DEFAULT;
  int immediate_display;
  int next_channel_delta;
  eqgui eg;
  int maxfd;
  int image_id = 0;

  eqgui_init(&eg);

  last_width = display_width;
  last_height = display_height;

  init(display_width, display_height);

  /* wait for main thread to signal channels are loaded */
  while (!x_can_go) {
    XEvent ev;
    XNextEvent(d, &ev);
    switch (ev.type) {
    case MapNotify:
    case ConfigureNotify:
      if (1) { //fullscreen == 0 /*&& wa.map_state == IsViewable*/) {
        Window _x;
        XWindowAttributes wa;
        XGetWindowAttributes(d, w, &wa);
        if (wa.map_state == IsViewable) {
          XTranslateCoordinates(d, w, wa.root, 0, 0, &wa.x, &wa.y, &_x);
          last_x = wa.x;
          last_y = wa.y;
          display_width = last_width = wa.width;
          display_height = last_height = wa.height;
          glClear(GL_COLOR_BUFFER_BIT);
        }
      }
      break;
    case Expose: display_image((id_image){ NULL, 0 }, 0, 0,
                               display_width, display_height, &eg);
    }
  }

restart:
  free(current_image.data);
  current_image = (image_item){ NULL, 0, 0 };
  glClear(GL_COLOR_BUFFER_BIT);
  glXSwapBuffers(d, w);

  image_number = 0;

  if (clock_gettime(CLOCK_MONOTONIC_RAW, &cur_time)) abort();
  cur_time_ms = cur_time.tv_sec * 1000 + cur_time.tv_nsec / 1000000;
  last_pointer_time = cur_time_ms;

  delay_ms = 0;

  maxfd = ConnectionNumber(d);
  if (maxfd < get_image_buffer_pipe()) maxfd = get_image_buffer_pipe();

  while (1) {
    next_image = get_image();
    image_id++;
    if (next_image.data != NULL) {
      free(current_image.data);
      current_image = next_image;
    }
    display_image((id_image){ current_image.data, image_id },
                  current_image.width, current_image.height,
                  display_width, display_height, &eg);
    if (next_image.data == NULL) goto no_image;
    if (clock_gettime(CLOCK_MONOTONIC_RAW, &cur_time)) abort();
    cur_time_ms = cur_time.tv_sec * 1000 + cur_time.tv_nsec / 1000000;
    /* let's hide the mouse pointer if it moved more than 500ms ago */
    if (cur_time_ms - last_pointer_time > 500 && pointer_visible &&
        gui_mode != GUI_MODE_EQ) {
      hide_pointer();
      pointer_visible = 0;
    }
    image_number++;
    /* we suppose 25 fps, so 40ms per image (fps should be read from stream) */
    //next_time_ms = world_start_time + image_number * 40;
    next_time_ms = get_start_time() + image_number * 40;
    delay_ms = next_time_ms - cur_time_ms;
    if (cur_time_ms > next_time_ms) delay_ms = 0;
    logprintf("computed delay %lu ms\n", delay_ms);
    /* we don't wait too much or too low */
    if (delay_ms > 40 * 2) delay_ms = 40 * 2;
    if (delay_ms < 40 / 2) delay_ms = 40 / 2;
    //usleep(delay_ms * 1000);
no_image:
    /* this event handling is a bit ugly */
    immediate_display = 0;
    next_channel_delta = 0;
    while (XPending(d)) {
      XEvent ev;
      XNextEvent(d, &ev);
      switch(ev.type) {
      case Expose: immediate_display = 1; break;
      case MapNotify:
      case ConfigureNotify:
        if (fullscreen == 0 /*&& wa.map_state == IsViewable*/) {
          Window _x;
          XWindowAttributes wa;
          XGetWindowAttributes(d, w, &wa);
          if (wa.map_state == IsViewable) {
            XTranslateCoordinates(d, w, wa.root, 0, 0, &wa.x, &wa.y, &_x);
            last_x = wa.x;
            last_y = wa.y;
            display_width = last_width = wa.width;
            display_height = last_height = wa.height;
            glClear(GL_COLOR_BUFFER_BIT);
          }
        } else {
          if (ev.type == ConfigureNotify) {
            display_width = ev.xconfigure.width;
            display_height = ev.xconfigure.height;
          }
        }
        break;
      case MotionNotify:
        if (gui_mode == GUI_MODE_EQ) {
          int ret = eqgui_pointer_move(&eg, ev.xmotion.x, ev.xmotion.y,
                             display_width, display_height);
          if (ret == EQGUI_REDRAW) immediate_display = 1;
          break;
        }
        /* fallthrought */
      case EnterNotify:
show_cursor:
        if (clock_gettime(CLOCK_MONOTONIC_RAW, &cur_time)) abort();
        if (pointer_visible == 0) show_pointer();
        last_pointer_time = cur_time.tv_sec * 1000 + cur_time.tv_nsec / 1000000;
        pointer_visible = 1;
        break;
      case ButtonRelease:
        if (gui_mode == GUI_MODE_EQ) {
          int ret = eqgui_button_release(&eg, ev.xbutton.x, ev.xbutton.y,
                             ev.xbutton.button, display_width, display_height);
          if (ret == EQGUI_REDRAW) immediate_display = 1;
        }
        break;
      case ButtonPress:
        if (gui_mode == GUI_MODE_EQ) {
          int ret = eqgui_button_press(&eg, ev.xbutton.x, ev.xbutton.y,
                             ev.xbutton.button, display_width, display_height);
          if (ret == EQGUI_REDRAW) immediate_display = 1;
          break;
        }
        if (gui_mode != GUI_MODE_DEFAULT) break;
        if (ev.xbutton.button == 5) goto volume_down;
        if (ev.xbutton.button == 4) goto volume_up;
        break;
      case KeyPress: {
        char buf[65];
        KeySym ks;
        XLookupString(&ev.xkey, buf, 64, &ks, NULL);
        if (gui_mode == GUI_MODE_EQ) {
          int ret = eqgui_key_press(&eg, ks);
          if (ret == EQGUI_REDRAW) { immediate_display = 1; break; }
          if (ret == EQGUI_QUIT) {
            gui_mode = GUI_MODE_DEFAULT;
            immediate_display = 1;
            goto show_cursor;
          }
          break;
        }
        if (gui_mode == GUI_MODE_GUIDE ||
            gui_mode == GUI_MODE_FULL_GUIDE ||
            gui_mode == GUI_MODE_HELP) {
          switch (ks) {
          case 'g': case 'G': case 'h': case XK_Escape:
            oclear();
            immediate_display = 1;
            gui_mode = GUI_MODE_DEFAULT;
            break;
          case XK_Up: case XK_KP_Up:
            scroll_guide(-1); immediate_display = 1; break;
          case XK_Down: case XK_KP_Down:
            scroll_guide(1); immediate_display = 1; break;
          case XK_Prior: case XK_KP_Prior:
            scroll_guide(-21); immediate_display = 1; break;
          case XK_Next: case XK_KP_Next: case ' ':
            scroll_guide(21); immediate_display = 1; break;
          case XK_Left:
            if (gui_mode == GUI_MODE_GUIDE)
              { next_guide(-1); immediate_display = 1; }
            break;
          case XK_Right:
            if (gui_mode == GUI_MODE_GUIDE)
              { next_guide(1); immediate_display = 1; }
            break;
          }
          break;
        }
        /* GUI_MODE_DEFAULT handler */
        switch (ks) {
        case 'q': exit(0);
        case 'E':
          gui_mode = GUI_MODE_EQ;
          eqgui_enter(&eg);
          immediate_display = 1;
          goto show_cursor;
        case 'f': toggle_fullscreen(last_x, last_y, last_width, last_height);
                  break;
        case 'F': if (current_image.data != NULL)
                    set_window_size_from_video(current_image.width,
                                               current_image.height);
                  break;
        case 'v':
        case XK_Down:
volume_down:
          mute = 0;
          volumeDB-=1;
          if (volumeDB<-60) volumeDB = -60;
          if (volumeDB==-60) volume = 0;
          else               volume = pow(10, volumeDB/20);
          oclear();
          oprintf("volume %g dB\n", volumeDB);
          odelay(500);
          break;
        case 'V':
        case XK_Up:
volume_up:
          mute = 0;
          volumeDB+=1;
          if (volumeDB>60) volumeDB = 60;
          volume = pow(10, volumeDB/20);
          oclear();
          oprintf("volume %g dB\n", volumeDB);
          odelay(500);
          break;
        case 'm':
          mute = 1 - mute;
          if (mute) volume = 0;
          else      volume = pow(10, volumeDB/20);
          oclear();
          if (mute) oprintf("mute");
          else      oprintf("volume %g dB\n", volumeDB);
          odelay(500);
          break;
        case 'M':
          mute = 0;
          volume = 1;
          volumeDB = 0;
          oclear();
          oprintf("volume %g dB\n", volumeDB);
          odelay(500);
          break;
        case 'S':
          audio_delay += 100;
          oclear();
          oprintf("audio start time %"PRId64" ms\n", audio_delay);
          odelay(500);
          break;
        case 's':
          audio_delay -= 100;
          oclear();
          oprintf("audio start time %"PRId64" ms\n", audio_delay);
          odelay(500);
          break;
        case 'a':
          if (current_image.data == NULL) break;
          current_aspect_ratio++;
          if (current_aspect_ratio >=
                  sizeof(aspect_ratios)/sizeof(aspect_ratios[0]))
            current_aspect_ratio = 0;
          if (current_aspect_ratio == 0) aspect_ratio_adjust = 1;
          else aspect_ratio_adjust = aspect_ratios[current_aspect_ratio] /
                                     ((float)current_image.width/
                                             current_image.height);
          oclear();
          oprintf("aspect_ratio %s\n",
                  aspect_ratios_name[current_aspect_ratio]);
          odelay(500);
          break;
        case 'g': {
          int get_current_channel(void);
          if (ev.xkey.state & ControlMask) { reload_guide(); break; }
          show_guide(get_current_channel());
          immediate_display = 1;
          gui_mode = GUI_MODE_GUIDE;
          break;
        }
        case 'G': {
          show_full_guide();
          immediate_display = 1;
          gui_mode = GUI_MODE_FULL_GUIDE;
          break;
        }
        case 'h': {
          show_help();
          immediate_display = 1;
          gui_mode = GUI_MODE_HELP;
          break;
        }
        case 'e':
          do_eq ^= 1;
          oclear();
          oprintf(do_eq ? "equalizer ON" : "equalizer OFF");
          odelay(500);
          break;
        case XK_Right: next_channel_delta++; break;
        case XK_Left: next_channel_delta--; break;
        case XK_KP_1: case XK_KP_End: case '1':
          if (set_channel(0)) goto restart; break;
        case XK_KP_2: case XK_KP_Down: case '2':
          if (set_channel(1)) goto restart; break;
        case XK_KP_3: case XK_KP_Next: case '3':
          if (set_channel(2)) goto restart; break;
        case XK_KP_4: case XK_KP_Left: case '4':
          if (set_channel(3)) goto restart; break;
        case XK_KP_5: case XK_KP_Begin: case '5':
          if (set_channel(4)) goto restart; break;
        case XK_KP_6: case XK_KP_Right: case '6':
          if (set_channel(5)) goto restart; break;
        case XK_KP_7: case XK_KP_Home: case '7':
          if (set_channel(6)) goto restart; break;
        case XK_KP_8: case XK_KP_Up: case '8':
          if (set_channel(7)) goto restart; break;
        case XK_KP_9: case XK_KP_Prior: case '9':
          if (set_channel(8)) goto restart; break;
        case XK_KP_0: case XK_KP_Insert: case '0':
          if (set_channel(9)) goto restart; break;
        }
        break;
      }}
    }
    if (next_channel_delta) { next_channel(next_channel_delta); goto restart; }
    if (immediate_display)
      display_image((id_image){ current_image.data, image_id },
                    current_image.width, current_image.height,
                    display_width, display_height, &eg);
    XFlush(d);
    if (clock_gettime(CLOCK_MONOTONIC_RAW, &cur_time)) abort();
    nt_ms = cur_time.tv_sec * 1000 + cur_time.tv_nsec / 1000000;
    if (nt_ms < cur_time_ms + delay_ms) {
      FD_ZERO(&rset);
      FD_SET(ConnectionNumber(d), &rset);
      timeout.tv_sec = 0;
      timeout.tv_usec = (cur_time_ms + delay_ms - nt_ms) * 1000;
      if (select(ConnectionNumber(d)+1, &rset, NULL, NULL, &timeout) == -1)
        abort();
      /* priority to X events */
      if (FD_ISSET(ConnectionNumber(d), &rset)) goto no_image;
    }
    FD_ZERO(&rset);
    FD_SET(ConnectionNumber(d), &rset);
    FD_SET(get_image_buffer_pipe(), &rset);
    if (select(maxfd+1, &rset, NULL, NULL, NULL) == -1)
      abort();
    /* at this point, putting image has priority over X */
    if (FD_ISSET(get_image_buffer_pipe(), &rset)) continue;
    goto no_image;
  }
  return NULL;
}

static int xerr(Display *d, XErrorEvent *e) { return 0; }
void force_x_redraw(void)
{
  /* let's go dirty... */
  Display *d = XOpenDisplay(0); if (d == NULL) abort();
  XEvent ev;
  /* I had a crash once where the window was created locally but not on the
   * server yet (XSendEvent failed)
   * so let's ignore errors by setting a dumb error handler
   */
  XSetErrorHandler(xerr);
  ev.type = Expose;
  ev.xexpose.display = d;
  ev.xexpose.window = w;
  ev.xexpose.x = 0;
  ev.xexpose.y = 0;
  ev.xexpose.width = -1;
  ev.xexpose.height = -1;
  XSendEvent(d, w, False, ExposureMask, &ev);
  XFlush(d);
  XCloseDisplay(d);
}
