#include <stdint.h>
#include <libxml/parser.h>
#include <string.h>
#include <inttypes.h>
#include "xml.h"
#include "logger.h"

void init_xml(void)
{
  xmlInitMemory();
  LIBXML_TEST_VERSION
}

static char *get_property(xmlAttrPtr attr, char *p)
{
  if (attr == NULL) return "";
  if (!strcmp((char*)attr->name, p)) return (char*)attr->children->content;
  return get_property(attr->next, p);
}

static uint64_t get_start(xmlNodePtr doc)
{
  char *p;
  if (doc == NULL) return 0;
  if (strcmp((char*)doc->name, "c") != 0) return get_start(doc->next);
  p = get_property(doc->properties, "t");
  if (p[0]) {
    uint64_t ret;
    if (sscanf(p, "%"SCNu64, &ret) != 1) return 0;
    return ret;
  }
  return get_start(doc->next);
}

static uint64_t get_inc(xmlNodePtr doc)
{
  char *p;
  if (doc == NULL) return 0;
  if (strcmp((char*)doc->name, "c") != 0) return get_inc(doc->next);
  p = get_property(doc->properties, "d");
  if (p[0]) {
    uint64_t ret;
    if (sscanf(p, "%"SCNu64, &ret) != 1) return 0;
    return ret;
  }
  return get_inc(doc->next);
}

static void get_best_quality(xmldata *d, xmlNodePtr doc)
{
  if (doc == NULL) return;
  if (!strcmp((char*)doc->name, "QualityLevel")) {
    uint64_t bitrate;
    sscanf(get_property(doc->properties, "Bitrate"), "%"SCNu64, &bitrate);
    if (bitrate > d->video_bitrate) {
      d->video_bitrate = bitrate;
      d->video_priv = get_property(doc->properties, "CodecPrivateData");
    }
  }
  get_best_quality(d, doc->next);
}

static void get_audio_quality(xmldata *d, xmlNodePtr doc)
{
  if (doc == NULL) return;
  if (!strcmp((char*)doc->name, "QualityLevel")) {
    uint64_t bitrate;
    sscanf(get_property(doc->properties, "Bitrate"), "%"SCNu64, &bitrate);
    if (bitrate > d->audio_bitrate) {
      d->audio_bitrate = bitrate;
      d->audio_priv = get_property(doc->properties, "CodecPrivateData");
    }
  }
  get_audio_quality(d, doc->next);
}

static void get_video(xmldata *d, xmlNodePtr doc)
{
  if (doc == NULL) return;
  if (!strcmp((char*)doc->name, "StreamIndex") &&
      !strcmp(get_property(doc->properties, "Type"), "video")) {
    d->video_url = get_property(doc->properties, "Url");
    sscanf(get_property(doc->properties, "TimeScale"), "%"SCNu64,
           &d->video_timescale);
    d->video_start = get_start(doc->children);
    d->video_inc = get_inc(doc->children);
    d->video_bitrate = 0;
    d->video_priv = "";
    get_best_quality(d, doc->children);
    return;
  }
  get_video(d, doc->children);
  get_video(d, doc->next);
}

static void get_audio(xmldata *d, xmlNodePtr doc)
{
  if (doc == NULL) return;
  if (!strcmp((char*)doc->name, "StreamIndex") &&
      !strcmp(get_property(doc->properties, "Type"), "audio")) {
    d->audio_url = get_property(doc->properties, "Url");
    sscanf(get_property(doc->properties, "TimeScale"), "%"SCNu64,
           &d->audio_timescale);
    d->audio_start = get_start(doc->children);
    d->audio_inc = get_inc(doc->children);
    d->audio_bitrate = 0;
    d->audio_priv = "";
    get_audio_quality(d, doc->children);
    return;
  }
  get_audio(d, doc->children);
  get_audio(d, doc->next);
}

int parse_xml(xmldata *d, char *data, int size)
{
  xmlDocPtr doc = xmlParseMemory(data, size);

  if (doc == NULL) { logprintf("no XML data?\n"); return -1; }

  get_video(d, doc->children);
  get_audio(d, doc->children);

  if (d->video_url != NULL) d->video_url = strdup(d->video_url);
  if (d->video_priv != NULL) d->video_priv = strdup(d->video_priv);
  //if (d->video_url == NULL || d->video_priv == NULL) abort();

  if (d->audio_url != NULL) d->audio_url = strdup(d->audio_url);
  if (d->audio_priv != NULL) d->audio_priv = strdup(d->audio_priv);
  //if (d->audio_url == NULL || d->audio_priv == NULL) abort();

  logprintf("video url %s\n", d->video_url);
  logprintf("video start %"PRIu64"\n", d->video_start);
  logprintf("video inc %"PRIu64"\n", d->video_inc);
  logprintf("video bitrate %"PRIu64"\n", d->video_bitrate);
  logprintf("video private data %s\n", d->video_priv);

  logprintf("audio url %s\n", d->audio_url);
  logprintf("audio start %"PRIu64"\n", d->audio_start);
  logprintf("audio inc %"PRIu64"\n", d->audio_inc);
  logprintf("audio bitrate %"PRIu64"\n", d->audio_bitrate);
  logprintf("audio private data %s\n", d->audio_priv);
  xmlFreeDoc(doc);

  return 0;
}
